/* 
 * dictionary.c --
 *
 *	This module implements a new object type named dictionary, and
 *	its associated Tcl commands. A dictionary has the same structure 
 *	as an array, but contrary to Tcl arrays it is a first-class 
 *	citizen: it can be passed by value, nested or stored in other 
 *	objects like lists. The main difference between arrays and 
 *	dictionaries is that arrays are variables, whereas dictionaries
 *	are data structures. Dictionaries can be used everywhere in place
 *	of arrays, only the syntax for variable access slightly differs.
 *
 * Copyright (c) 1999-2000 Frdric BONNET <frederic.bonnet@ciril.fr>
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "dictionary.h"
#include "dictionaryInt.h"

#ifndef USE_TCL_STUBS
#undef Tcl_InitStubs
#define Tcl_InitStubs(a,b,c) Tcl_PkgRequire(a,"Tcl",b,c)
#endif

/* 
 * Reference to Tcl list type. Used in type conversion code.
 * Initialized in Dictionary_Init().
 */

Tcl_ObjType* listObjTypePtr;




/*
 *--------------------------------------------------------------
 *
 * DictionaryCreateObjCmd --
 *
 *	This procedure is invoked to process the "create"
 *	subcommand for the Tcl command "dictionary". It creates
 *	a new dictionary object.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Some objects type may change.
 *
 *--------------------------------------------------------------
 */

int
DictionaryCreateObjCmd(clientData, interp, objc, objv)
    ClientData	clientData;
    Tcl_Interp	*interp;
    int		objc;
    Tcl_Obj	*CONST objv[];
{
    int listc;
    Tcl_Obj **listv;

    if (objc == 3) {
	/* 
	 * Initial value is provided as a single object 
	 */

	if (objv[2]->typePtr == &dictionaryObjType) {
	    /* Object is already a dictionary, return it */
	    Tcl_SetObjResult(interp, objv[2]);
	    return TCL_OK;
	}

	/* Try to convert argument to a proper list */
	if (Tcl_ListObjGetElements(interp, objv[2], &listc, &listv)
	    == TCL_ERROR) {
	    return TCL_ERROR;
	}
    } else {
	listc = objc-2;
	listv = (Tcl_Obj **) objv+2;
    }

    /* List must have an even number of elements */
    if (listc % 2 != 0) {
	goto oddListError;
    }

    /* Create the new object */
    Tcl_SetObjResult(interp, Dict_NewDictionaryObj(listc, listv));
    return TCL_OK;

    /*
     * Error handling
     */

oddListError:
    Tcl_AppendResult(interp, 
		     "list must have an even number of elements", 
		     NULL);
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * DictionarySetObjCmd --
 *
 *	This procedure is invoked to process the "set"
 *	subcommand for the Tcl command "dictionary". It sets
 *	the value of a dictionary element.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The given variable value is altered. Some objects type 
 *	may change.
 *
 *--------------------------------------------------------------
 */

int
DictionarySetObjCmd(clientData, interp, objc, objv)
    ClientData	clientData;
    Tcl_Interp	*interp;
    int		objc;
    Tcl_Obj	*CONST objv[];
{
    Tcl_Obj *dictPtr;

    /* 
     * Sanity check
     */

    if (objc != 4 && objc != 5) {
	goto numberArgsError;
    }

    dictPtr = Tcl_ObjGetVar2(interp, objv[2], (Tcl_Obj *) NULL,
				  (TCL_LEAVE_ERR_MSG));
    if (dictPtr == NULL) {
	/* Variable doesn't exist */
	return TCL_ERROR;
    }

    if (objc == 4) {

	/* 
	 * Query value 
	 */

	Tcl_Obj *valuePtr;

	if (Dict_DictionaryObjGetElement(interp, dictPtr, 
					Tcl_GetString(objv[3]), &valuePtr)
	    == TCL_ERROR) {
	    return TCL_ERROR;
	}
	if (valuePtr == NULL) {
	    goto noSuchKeyError;
	}
	Tcl_SetObjResult(interp, valuePtr);
	return TCL_OK;

    } else {
	
	/* 
	 * Set value 
	 */

	int newElement;
	int shared;

	shared = Tcl_IsShared(dictPtr);

	/*
	 * If object is shared, make a copy
	 */

	if (shared) {
	    dictPtr = Tcl_DuplicateObj(dictPtr);
	    Tcl_IncrRefCount(dictPtr);
	}
	if (Dict_DictionaryObjSetElement(interp, dictPtr, 
					Tcl_GetString(objv[3]), objv[4], 
					&newElement)
	    == TCL_ERROR) {
	    if (shared) {
		Tcl_DecrRefCount(dictPtr);
	    }
	    return TCL_ERROR;
	}

	/*
	 * If object was shared, set the variable value back
	 */

	if (shared) {
	    if (Tcl_ObjSetVar2(interp, objv[2], (Tcl_Obj *) NULL, dictPtr,
			       (TCL_LEAVE_ERR_MSG))
		== NULL) {
		Tcl_DecrRefCount(dictPtr);
		return TCL_ERROR;
	    }
	}
	Tcl_SetObjResult(interp, objv[4]);
	return TCL_OK;
    }

    /* 
     * Error handling
     */

numberArgsError:
    Tcl_WrongNumArgs(interp, 2, objv, "varName keyName ?newValue?");
    return TCL_ERROR;

noSuchKeyError:
    Tcl_AppendResult(interp, 
		     "no such key: \"", Tcl_GetString(objv[3]), "\"",
		     NULL);
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * DictionaryUnsetObjCmd --
 *
 *	This procedure is invoked to process the "unset"
 *	subcommand for the Tcl command "dictionary". It unsets
 *	one or several dictionary elements.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The given variable value is altered. Some objects type 
 *	may change.
 *
 *--------------------------------------------------------------
 */

int
DictionaryUnsetObjCmd(clientData, interp, objc, objv)
    ClientData	clientData;
    Tcl_Interp	*interp;
    int		objc;
    Tcl_Obj	*CONST objv[];
{
    Tcl_Obj *dictPtr;
    int i, shared;

    /* 
     * Sanity check
     */

    if (objc < 4) {
	goto numberArgsError;
    }

    dictPtr = Tcl_ObjGetVar2(interp, objv[2], (Tcl_Obj *) NULL,
				  (TCL_LEAVE_ERR_MSG));
    if (dictPtr == NULL) {
	/* Variable doesn't exist */
	return TCL_ERROR;
    }

    shared = Tcl_IsShared(dictPtr);

    /*
     * If object is shared, make a copy
     */

    if (shared) {
	dictPtr = Tcl_DuplicateObj(dictPtr);
	Tcl_IncrRefCount(dictPtr);
    }

    for (i=3; i<objc; i++) {
	if (Dict_DictionaryObjUnsetElement(interp, dictPtr, 
					  Tcl_GetString(objv[i]))
	    == TCL_ERROR) {
	    if (shared) {
		Tcl_DecrRefCount(dictPtr);
	    }
	    return TCL_ERROR;
	}
    }

    /*
     * If object was shared, set the variable value back
     */

    if (shared) {
	if (Tcl_ObjSetVar2(interp, objv[2], (Tcl_Obj *) NULL, dictPtr,
			   (TCL_LEAVE_ERR_MSG))
	    == NULL) {
	    Tcl_DecrRefCount(dictPtr);
	    return TCL_ERROR;
	}
    }
    return TCL_OK;

    /* 
     * Error handling
     */

numberArgsError:
    Tcl_WrongNumArgs(interp, 2, objv, "varName keyName ?keyName ...?");
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * DictionarySizeObjCmd --
 *
 *	This procedure is invoked to process the "size"
 *	subcommand for the Tcl command "dictionary". It returns
 *	the integer size of the specified dictionary object.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The specified object type may change.
 *
 *--------------------------------------------------------------
 */

int
DictionarySizeObjCmd(clientData, interp, objc, objv)
    ClientData	clientData;
    Tcl_Interp	*interp;
    int		objc;
    Tcl_Obj	*CONST objv[];
{
    int size;

    /* 
     * Sanity check
     */

    if (objc != 3) {
	goto numberArgsError;
    }

    if (Dict_DictionaryObjGetSize(interp, objv[2], &size)
	== TCL_ERROR) {
	return TCL_ERROR;
    }

    Tcl_SetObjResult(interp, Tcl_NewIntObj(size));
    return TCL_OK;
    
    /* 
     * Error handling
     */

numberArgsError:
    Tcl_WrongNumArgs(interp, 2, objv, "dictionary");
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * DictionaryExistsObjCmd --
 *
 *	This procedure is invoked to process the "exists"
 *	subcommand for the Tcl command "dictionary". It returns
 *	whether the given key exists in the specified dictionary.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The specified object type may change.
 *
 *--------------------------------------------------------------
 */

int
DictionaryExistsObjCmd(clientData, interp, objc, objv)
    ClientData	clientData;
    Tcl_Interp	*interp;
    int		objc;
    Tcl_Obj	*CONST objv[];
{
    Tcl_Obj *valuePtr;

    /* 
     * Sanity check
     */

    if (objc != 4) {
	goto numberArgsError;
    }

    if (Dict_DictionaryObjGetElement(interp, objv[2], Tcl_GetString(objv[3]), 
				    &valuePtr)
	== TCL_ERROR) {
	return TCL_ERROR;
    }

    Tcl_SetObjResult(interp, Tcl_NewBooleanObj(valuePtr != NULL));
    return TCL_OK;
    
    /* 
     * Error handling
     */

numberArgsError:
    Tcl_WrongNumArgs(interp, 2, objv, "dictionary keyName");
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * DictionaryAddObjCmd --
 *
 *	This procedure is invoked to process the "add"
 *	subcommand for the Tcl command "dictionary". It adds
 *	one or several elements to a dictionary variable from
 *	a specified list or dictionary.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The given variable value is altered. Some objects type 
 *	may change.
 *
 *--------------------------------------------------------------
 */

int
DictionaryAddObjCmd(clientData, interp, objc, objv)
    ClientData	clientData;
    Tcl_Interp	*interp;
    int		objc;
    Tcl_Obj	*CONST objv[];
{
    register Tcl_Obj *dictPtr;
    int shared;
    int i, listc, newKey;
    Tcl_Obj **listv;

    /* 
     * Sanity check
     */

    if (objc < 4) {
	goto numberArgsError;
    }

    dictPtr = Tcl_ObjGetVar2(interp, objv[2], (Tcl_Obj *) NULL,
				  (TCL_LEAVE_ERR_MSG));
    if (dictPtr == NULL) {
	/* Variable doesn't exist */
	return TCL_ERROR;
    }

    /*
     * Attempt to convert object to a dictionary
     */

    if (Tcl_ConvertToType(interp, dictPtr, &dictionaryObjType)
	== TCL_ERROR) {
	return TCL_ERROR;
    }

    if (objc == 4) {

	/* 
	 * Add elements from one list or dictionary argument
	 */

	if (objv[3]->typePtr == &dictionaryObjType) {

	    /* 
	     * If the specified argument is a dictionary
	     */

	    Tcl_HashTable *table1Ptr, *table2Ptr;
	    Tcl_HashSearch search;
	    Tcl_HashEntry *entryPtr;

	    Dict_DictionaryObjGetElements(interp, dictPtr, &table1Ptr);
	    Dict_DictionaryObjGetElements(interp, objv[3], &table2Ptr);

	    /*
	     * Quick test: if objects share the same internals, do nothing.
	     */

	    if (table1Ptr == table2Ptr) {
		return TCL_OK;
	    }

	    /*
	     * Add all elements of dictionary 2 to dictionary 1
	     */

	    shared = Tcl_IsShared(dictPtr);

	    /*
	     * If object is shared, make a copy
	     */

	    if (shared) {
		dictPtr = Tcl_DuplicateObj(dictPtr);
		Tcl_IncrRefCount(dictPtr);
	    }

	    /*
	     * Add elements to dictionary
	     */

	    for (entryPtr = Tcl_FirstHashEntry(table2Ptr, &search);
		 entryPtr != (Tcl_HashEntry*) NULL;
		 entryPtr = Tcl_NextHashEntry(&search)) {

		if (Dict_DictionaryObjSetElement(interp, dictPtr, 
						Tcl_GetHashKey(table2Ptr, entryPtr),
						(Tcl_Obj *) Tcl_GetHashValue(entryPtr),
						&newKey)
		    == TCL_ERROR) {

		    if (shared) {
			Tcl_DecrRefCount(dictPtr);
		    }
		    return TCL_ERROR;
		}
	    }

	    /*
	     * If object was shared, set the variable value back
	     */

	    if (shared) {
		if (Tcl_ObjSetVar2(interp, objv[2], (Tcl_Obj *) NULL, 
				   dictPtr, (TCL_LEAVE_ERR_MSG))
		    == NULL) {
		    Tcl_DecrRefCount(dictPtr);
		    return TCL_ERROR;
		}
		Tcl_DecrRefCount(dictPtr);
	    }
	    return TCL_OK;

	}

	/* Try to convert argument to a proper list */
	if (Tcl_ListObjGetElements(interp, objv[3], &listc, &listv)
	    == TCL_ERROR) {
	    return TCL_ERROR;
	}
    } else {
	listc = objc-3;
	listv = (Tcl_Obj **) objv+3;
    }

    /* List must have an even number of elements */
    if (listc % 2 != 0) {
	goto oddListError;
    }

    if (listc == 0) {
	/* Nothing to do */
	return TCL_OK;
    }

    /*
     * Add all elements of list to dictionary
     */

    shared = Tcl_IsShared(dictPtr);

    /*
     * If object is shared, make a copy
     */

    if (shared) {
	dictPtr = Tcl_DuplicateObj(dictPtr);
	Tcl_IncrRefCount(dictPtr);
    }

    for (i=0; i<listc; i+=2) {
	if (Dict_DictionaryObjSetElement(interp, dictPtr, 
					Tcl_GetString(listv[i]),
					listv[i+1],
					&newKey)
	    == TCL_ERROR) {
	    if (shared) {
		Tcl_DecrRefCount(dictPtr);
	    }
	    return TCL_ERROR;
	}
    }

    /*
     * If object was shared, set the variable value back
     */

    if (shared) {
	if (Tcl_ObjSetVar2(interp, objv[2], (Tcl_Obj *) NULL, 
			   dictPtr, (TCL_LEAVE_ERR_MSG))
	    == NULL) {
	    Tcl_DecrRefCount(dictPtr);
	    return TCL_ERROR;
	}
	Tcl_DecrRefCount(dictPtr);
    }
    return TCL_OK;

    /* 
     * Error handling
     */

numberArgsError:
    Tcl_WrongNumArgs(interp, 2, objv, "varName key value ?key value ...?");
    Tcl_WrongNumArgs(interp, 2, objv, "varName list");
    return TCL_ERROR;

oddListError:
    Tcl_AppendResult(interp, 
		     "list must have an even number of elements", 
		     NULL);
    return TCL_ERROR;
}



/*
 *--------------------------------------------------------------
 *
 * DictionaryMergeObjCmd --
 *
 *	This procedure is invoked to process the "add"
 *	subcommand for the Tcl command "dictionary". It adds
 *	one or several elements to a dictionary from a specified 
 *	list or dictionary, and return a new dictionary object.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Some objects type may change.
 *
 *--------------------------------------------------------------
 */

int
DictionaryMergeObjCmd(clientData, interp, objc, objv)
    ClientData	clientData;
    Tcl_Interp	*interp;
    int		objc;
    Tcl_Obj	*CONST objv[];
{
    register Tcl_Obj *dictPtr;
    int i, listc, newKey;
    Tcl_Obj **listv;

    /* 
     * Sanity check
     */

    if (objc < 4) {
	goto numberArgsError;
    }

    /*
     * Attempt to convert object to a dictionary
     */

    if (Tcl_ConvertToType(interp, objv[2], &dictionaryObjType)
	== TCL_ERROR) {
	return TCL_ERROR;
    }

    if (objc == 4) {

	/* 
	 * Add elements from one list or dictionary argument
	 */

	if (objv[3]->typePtr == &dictionaryObjType) {

	    /* 
	     * If the specified argument is a dictionary
	     */

	    Tcl_HashTable *table1Ptr, *table2Ptr;
	    Tcl_HashSearch search;
	    Tcl_HashEntry *entryPtr;

	    Dict_DictionaryObjGetElements(interp, objv[2], &table1Ptr);
	    Dict_DictionaryObjGetElements(interp, objv[3], &table2Ptr);

	    /*
	     * Quick tests
	     */

	    if (table1Ptr == table2Ptr) {
		/* Objects are the same, return it */
		Tcl_SetObjResult(interp, objv[2]);
		return TCL_OK;
	    } else if (table1Ptr->numEntries == 0) {
		/* First object is empty, return second */
		Tcl_SetObjResult(interp, objv[3]);
		return TCL_OK;
	    } else if (table2Ptr->numEntries == 0) {
		/* Second object is empty, return first */
		Tcl_SetObjResult(interp, objv[2]);
		return TCL_OK;
	    }

	    /*
	     * Create a new dictionary and add elements
	     */

	    dictPtr = Tcl_DuplicateObj(objv[2]);
	    Tcl_IncrRefCount(dictPtr);

	    for (entryPtr = Tcl_FirstHashEntry(table2Ptr, &search);
		 entryPtr != (Tcl_HashEntry*) NULL;
		 entryPtr = Tcl_NextHashEntry(&search)) {

		if (Dict_DictionaryObjSetElement(interp, dictPtr, 
						Tcl_GetHashKey(table2Ptr, entryPtr),
						(Tcl_Obj *) Tcl_GetHashValue(entryPtr),
						&newKey)
		    == TCL_ERROR) {

		    Tcl_DecrRefCount(dictPtr);
		    return TCL_ERROR;
		}
	    }

	    Tcl_SetObjResult(interp, dictPtr);
	    Tcl_DecrRefCount(dictPtr);
	    return TCL_OK;
	}

	/* Try to convert argument to a proper list */
	if (Tcl_ListObjGetElements(interp, objv[3], &listc, &listv)
	    == TCL_ERROR) {
	    return TCL_ERROR;
	}
    } else {
	listc = objc-3;
	listv = (Tcl_Obj **) objv+3;
    }

    /* List must have an even number of elements */
    if (listc % 2 != 0) {
	goto oddListError;
    }

    if (listc == 0) {
	/* Second object is empty, return first */
	Tcl_SetObjResult(interp, objv[2]);
	return TCL_OK;
    }

    /*
     * Create a new dictionary and add elements
     */

    dictPtr = Tcl_DuplicateObj(objv[2]);
    Tcl_IncrRefCount(dictPtr);

    for (i=0; i<listc; i+=2) {
	if (Dict_DictionaryObjSetElement(interp, dictPtr, 
					Tcl_GetString(listv[i]),
					listv[i+1],
					&newKey)
	    == TCL_ERROR) {
	    Tcl_DecrRefCount(dictPtr);
	    return TCL_ERROR;
	}
    }

    Tcl_SetObjResult(interp, dictPtr);
    Tcl_DecrRefCount(dictPtr);
    return TCL_OK;

    /* 
     * Error handling
     */

numberArgsError:
    Tcl_WrongNumArgs(interp, 2, objv, "dictionary ( list | key value ?key value ...? )");
    return TCL_ERROR;

oddListError:
    Tcl_AppendResult(interp, 
		     "list must have an even number of elements", 
		     NULL);
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * DictionarySubtractObjCmd --
 *
 *	This procedure is invoked to process the "subtract"
 *	subcommand for the Tcl command "dictionary". It subtract
 *	one or several elements in a dictionary from a specified 
 *	list or dictionary, and return a new dictionary object.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Some objects type may change.
 *
 *--------------------------------------------------------------
 */

int
DictionarySubtractObjCmd(clientData, interp, objc, objv)
    ClientData	clientData;
    Tcl_Interp	*interp;
    int		objc;
    Tcl_Obj	*CONST objv[];
{
    register Tcl_Obj *dictPtr;
    int i, listc;
    Tcl_Obj **listv;
    int listIncr = 1;

    /* 
     * Sanity check
     */

    if (objc < 4) {
	goto numberArgsError;
    }

    /*
     * Attempt to convert object to a dictionary
     */

    if (Tcl_ConvertToType(interp, objv[2], &dictionaryObjType)
	== TCL_ERROR) {
	return TCL_ERROR;
    }

    if (objc == 4) {

	/* 
	 * Subtract elements from one list or dictionary argument
	 */

	if (objv[3]->typePtr == &dictionaryObjType) {

	    /* 
	     * If the specified argument is a dictionary
	     */

	    Tcl_HashTable *table1Ptr, *table2Ptr;
	    Tcl_HashSearch search;
	    Tcl_HashEntry *entryPtr;

	    Dict_DictionaryObjGetElements(interp, objv[2], &table1Ptr);
	    Dict_DictionaryObjGetElements(interp, objv[3], &table2Ptr);

	    /*
	     * Quick tests
	     */

	    if (table1Ptr == table2Ptr) {
		/* Objects are the same, return an empty result */
		Tcl_ResetResult(interp);
		return TCL_OK;
	    } else if (table1Ptr->numEntries == 0) {
		/* First object is empty, return an empty result */
		Tcl_ResetResult(interp);
		return TCL_OK;
	    } else if (table2Ptr->numEntries == 0) {
		/* Second object is empty, return first object */
		Tcl_SetObjResult(interp, objv[2]);
		return TCL_OK;
	    }

	    /*
	     * Create a new dictionary and subtract elements
	     */

	    dictPtr = Tcl_DuplicateObj(objv[2]);
	    Tcl_IncrRefCount(dictPtr);

	    for (entryPtr = Tcl_FirstHashEntry(table2Ptr, &search);
		 entryPtr != (Tcl_HashEntry*) NULL;
		 entryPtr = Tcl_NextHashEntry(&search)) {

		/* 
		 * Do not report errors due to missing elements: interp is NULL
		 */

		Dict_DictionaryObjUnsetElement(NULL, dictPtr, 
					       Tcl_GetHashKey(table2Ptr, entryPtr));
	    }

	    Tcl_SetObjResult(interp, dictPtr);
	    Tcl_DecrRefCount(dictPtr);
	    return TCL_OK;
	}

	/* Try to convert argument to a proper list */
	if (Tcl_ListObjGetElements(interp, objv[3], &listc, &listv)
	    == TCL_ERROR) {
	    return TCL_ERROR;
	}

	/* List must have an even number of elements */
	if (listc % 2 != 0) {
	    goto oddListError;
	}
	listIncr = 2;	/* Keys are even list elements */

    } else {
	listc = objc-3;
	listv = (Tcl_Obj **) objv+3;

	listIncr = 1;	/* Keys are all list elements */
    }

    if (listc == 0) {
	/* Second object is empty, return first */
	Tcl_SetObjResult(interp, objv[2]);
	return TCL_OK;
    }

    /*
     * Create a new dictionary and subtract elements
     */

    dictPtr = Tcl_DuplicateObj(objv[2]);
    Tcl_IncrRefCount(dictPtr);

    for (i=0; i<listc; i+=listIncr) {

	/* 
	 * Do not report errors due to missing elements: interp is NULL
	 */

	Dict_DictionaryObjUnsetElement(NULL, dictPtr, 
				       Tcl_GetString(listv[i]));
    }

    Tcl_SetObjResult(interp, dictPtr);
    Tcl_DecrRefCount(dictPtr);
    return TCL_OK;

    /* 
     * Error handling
     */

numberArgsError:
    Tcl_WrongNumArgs(interp, 2, objv, "dictionary ( list | key ?key ...? )");
    return TCL_ERROR;

oddListError:
    Tcl_AppendResult(interp, 
		     "list must have an even number of elements", 
		     NULL);
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * DictionaryGetObjCmd --
 *
 *	This procedure is invoked to process the "get"
 *	subcommand for the Tcl command "dictionary". It queries
 *	one or several dictionary elements whose keys match the
 *	given pattern. It returns a new dictionary.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The object type may change.
 *
 *--------------------------------------------------------------
 */

int
DictionaryGetObjCmd(clientData, interp, objc, objv)
    ClientData	clientData;
    Tcl_Interp	*interp;
    int		objc;
    Tcl_Obj	*CONST objv[];
{
    Tcl_HashTable *tablePtr;
    Tcl_HashSearch search;
    Tcl_HashEntry *entryPtr;
    Tcl_Obj *dictPtr;
    char *key, *pattern;
    int newKey;

    /* 
     * Sanity check
     */

    if (objc != 4) {
	goto numberArgsError;
    }

    if (Dict_DictionaryObjGetElements(interp, objv[2], &tablePtr)
	== TCL_ERROR) {
	return TCL_ERROR;
    }
    pattern = Tcl_GetString(objv[3]);

    dictPtr = Dict_NewDictionaryObj(0, NULL);
    Tcl_IncrRefCount(dictPtr);

    for (entryPtr = Tcl_FirstHashEntry(tablePtr, &search);
	 entryPtr != (Tcl_HashEntry*) NULL;
	 entryPtr = Tcl_NextHashEntry(&search)) {

	key = Tcl_GetHashKey(tablePtr, entryPtr);
	if (!Tcl_StringMatch(key, pattern)) {
	    continue;
	}
	if (Dict_DictionaryObjSetElement(interp, dictPtr, 
					key,
					(Tcl_Obj *) Tcl_GetHashValue(entryPtr),
					&newKey)
	    == TCL_ERROR) {
	    Tcl_DecrRefCount(dictPtr);
	    return TCL_ERROR;
	}
    }

    Tcl_SetObjResult(interp, dictPtr);
    Tcl_DecrRefCount(dictPtr);
    return TCL_OK;

    /* 
     * Error handling
     */

numberArgsError:
    Tcl_WrongNumArgs(interp, 2, objv, "dictionary pattern");
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * DictionaryNamesObjCmd --
 *
 *	This procedure is invoked to process the "names"
 *	subcommand for the Tcl command "dictionary". It queries
 *	the dictionary keys that match the given pattern.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The object type may change.
 *
 *--------------------------------------------------------------
 */

int
DictionaryNamesObjCmd(clientData, interp, objc, objv)
    ClientData	clientData;
    Tcl_Interp	*interp;
    int		objc;
    Tcl_Obj	*CONST objv[];
{
    Tcl_HashTable *tablePtr;
    Tcl_HashSearch search;
    Tcl_HashEntry *entryPtr;
    Tcl_Obj *resultPtr;
    char *key, *pattern = NULL;

    /* 
     * Sanity check
     */

    if (objc < 3 || objc > 4) {
	goto numberArgsError;
    }

    if (Dict_DictionaryObjGetElements(interp, objv[2], &tablePtr)
	== TCL_ERROR) {
	return TCL_ERROR;
    }
    if (objc >= 4) {
	pattern = Tcl_GetString(objv[3]);
    }

    resultPtr = Tcl_NewObj();
    Tcl_IncrRefCount(resultPtr);

    for (entryPtr = Tcl_FirstHashEntry(tablePtr, &search);
	 entryPtr != (Tcl_HashEntry*) NULL;
	 entryPtr = Tcl_NextHashEntry(&search)) {

	key = Tcl_GetHashKey(tablePtr, entryPtr);
	if (pattern && !Tcl_StringMatch(key, pattern)) {
	    continue;
	}
	if (Tcl_ListObjAppendElement(interp, resultPtr, 
				     Tcl_NewStringObj(key, -1))
	    == TCL_ERROR) {
	    Tcl_DecrRefCount(resultPtr);
	    return TCL_ERROR;
	}
    }

    Tcl_SetObjResult(interp, resultPtr);
    Tcl_DecrRefCount(resultPtr);
    return TCL_OK;

    /* 
     * Error handling
     */

numberArgsError:
    Tcl_WrongNumArgs(interp, 2, objv, "dictionary ?pattern?");
    return TCL_ERROR;
}




/*
 *--------------------------------------------------------------
 *
 * DictionaryForeachObjCmd --
 *
 *	This procedure is invoked to process the "foreach"
 *	subcommand for the Tcl command "dictionary". It iterates
 *	over one or several dictionaries' elements.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The object type may change.
 *
 *--------------------------------------------------------------
 */

int
DictionaryForeachObjCmd(clientData, interp, objc, objv)
    ClientData	clientData;
    Tcl_Interp	*interp;
    int		objc;
    Tcl_Obj	*CONST objv[];
{
    int result = TCL_OK;
    int numDicts;
    int maxSize;
    Tcl_Obj *bodyPtr;

#define STATIC_DICT_SIZE 4
    Tcl_Obj *keyVarNameArray[STATIC_DICT_SIZE]; /* Array of key var names */
    Tcl_Obj *valVarNameArray[STATIC_DICT_SIZE]; /* Array of value var names */
    Tcl_Obj *dictArray[STATIC_DICT_SIZE];
    Tcl_HashTable *tableArray[STATIC_DICT_SIZE];
    Tcl_HashEntry *entryArray[STATIC_DICT_SIZE];
    Tcl_HashSearch searchArray[STATIC_DICT_SIZE];

    Tcl_Obj **keyVarName = keyVarNameArray;
    Tcl_Obj **valVarName = valVarNameArray;
    Tcl_Obj **dict = dictArray;
    Tcl_HashTable **table = tableArray;
    Tcl_HashEntry **entry = entryArray;
    Tcl_HashSearch *search = searchArray;


    int i,j,res;
    int vc;
    Tcl_Obj **vv;
    int size;
    Tcl_HashTable *t;
    Tcl_Obj *key, *value;

    /* 
     * Sanity check
     */

    if (objc < 5 || (objc % 2) != 1) {
	goto numberArgsError;
    }

    /*
     * Parse command line
     */

    numDicts = (objc-3)/2;
    if (numDicts > STATIC_DICT_SIZE) {
	keyVarName = (Tcl_Obj **) Tcl_Alloc(numDicts * sizeof(Tcl_Obj *));
	valVarName = (Tcl_Obj **) Tcl_Alloc(numDicts * sizeof(Tcl_Obj *));
	dict = (Tcl_Obj **) Tcl_Alloc(numDicts * sizeof(Tcl_Obj *));
	table = (Tcl_HashTable **) Tcl_Alloc(numDicts * sizeof(Tcl_HashTable *));
	entry = (Tcl_HashEntry **) Tcl_Alloc(numDicts * sizeof(Tcl_HashEntry *));
	search = (Tcl_HashSearch *) Tcl_Alloc(numDicts * sizeof(Tcl_HashSearch));

    }
    for (i=0; i<numDicts; i++) {
	keyVarName[i] = NULL;
	valVarName[i] = NULL;
	dict[i]       = NULL;
    }

    maxSize = 0;
    for (i=0; i<numDicts; i++) {
	result = Tcl_ListObjGetElements(interp, objv[2+i*2], &vc, &vv);
	if (result != TCL_OK) {
	    goto cleanup;
	}
	if (vc != 2) {
	    Tcl_SetResult(interp, "list of variables must have two elements",
			  TCL_STATIC);
	    result = TCL_ERROR;
	    goto cleanup;
	}
	keyVarName[i] = vv[0]; Tcl_IncrRefCount(keyVarName[i]);
	valVarName[i] = vv[1]; Tcl_IncrRefCount(valVarName[i]);
	dict[i] = objv[3+i*2]; Tcl_IncrRefCount(dict[i]);

	result = Dict_DictionaryObjGetSize(interp, dict[i], &size);
	if (result != TCL_OK) {
	    goto cleanup;
	}
	if (maxSize < size) {
	    maxSize = size;
	}
    }

    /*
     * Store hash table info before iterating. We'll stop iterating over
     * dictionaries whose internals are modified.
     */

    for (i=0; i<numDicts; i++) {
	Dict_DictionaryObjGetElements(interp, dict[i], &table[i]);
	entry[i] = Tcl_FirstHashEntry(table[i], &search[i]);

    }

    /*
     * Iterate maxSize times through the dictionaries in parallel.
     * If some value dictionary run out of values, set loop vars to ""
     */

    bodyPtr = objv[objc-1];
    for (j=0; j<maxSize; j++) {

	/*
	 * Set key and value variables
	 */

	for (i=0; i<numDicts; i++) {
	    result = Dict_DictionaryObjGetElements(interp, dict[i], &t);
	    if (result != TCL_OK) {
		goto cleanup;
	    }
	    if (   t != table[i]	/* Dictionary has been modified */
		|| entry[i] == NULL	/* No more elements */
	       ) {
		key   = Tcl_NewObj();
		value = Tcl_NewObj();
	    } else {
		key = Tcl_NewStringObj(Tcl_GetHashKey(table[i], entry[i]), -1);
		value = (Tcl_Obj *) Tcl_GetHashValue(entry[i]);

		/* Get next entry */
		entry[i] = Tcl_NextHashEntry(&search[i]);
	    }

	    /* Set variables */
	    Tcl_IncrRefCount(key);
	    Tcl_IncrRefCount(value);
	    res = (   Tcl_ObjSetVar2(interp, keyVarName[i], NULL, 
				     key,
				     TCL_LEAVE_ERR_MSG)
		      != NULL
		   && Tcl_ObjSetVar2(interp, valVarName[i], NULL, 
				     value,
				     TCL_LEAVE_ERR_MSG)
		      != NULL
		  );
	    Tcl_DecrRefCount(key);
	    Tcl_DecrRefCount(value);
	    if (!res) {
		result = TCL_ERROR;
		goto cleanup;
	    }
	}

	/*
	 * Evaluate script
	 */
	
	result = Tcl_EvalObjEx(interp, bodyPtr, 0);
	if (result != TCL_OK) {
	    if (result == TCL_CONTINUE) {
		result = TCL_OK;
	    } else if (result == TCL_BREAK) {
		result = TCL_OK;
		break;
	    } else if (result == TCL_ERROR) {
                char msg[32 + TCL_INTEGER_SPACE];

		sprintf(msg, "\n    (\"foreach\" body line %d)",
			interp->errorLine);
		Tcl_AddObjErrorInfo(interp, msg, -1);
		break;
	    } else {
		break;
	    }
	}
    }
    if (result == TCL_OK) {
	Tcl_ResetResult(interp);
    }     

cleanup:
    for (i=0; i<numDicts; i++) {
	if (keyVarName[i]) Tcl_DecrRefCount(keyVarName[i]);
	if (valVarName[i]) Tcl_DecrRefCount(valVarName[i]);
	if (dict[i])       Tcl_DecrRefCount(dict[i]);
    }
    if (numDicts > STATIC_DICT_SIZE) {
	Tcl_Free((char *) keyVarName);
	Tcl_Free((char *) valVarName);
	Tcl_Free((char *) dict);
	Tcl_Free((char *) table);
	Tcl_Free((char *) entry);
	Tcl_Free((char *) search);
    }
    return result;

    /* 
     * Error handling
     */

numberArgsError:
    Tcl_WrongNumArgs(interp, 2, objv, "{keyVar valueVar} dictionary ?{keyVar valueVar} dictionary ...? command");
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * DictionaryObjCmd --
 *
 *	This procedure is invoked to process the Tcl command
 *	"dictionary", and serves as a switchbox between the
 *	different subcommands.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Depend on subcommand.
 *
 *--------------------------------------------------------------
 */

int
DictionaryObjCmd(clientData, interp, objc, objv)
    ClientData	clientData;
    Tcl_Interp	*interp;
    int		objc;
    Tcl_Obj	*CONST objv[];
{
    /*
     * The list of constants below should match the options string array
     * and commands function array below.
     */

    enum {
	option_add,     option_array, option_create,   option_exists,
	option_foreach, option_get,   option_merge,    option_name,
	option_set,     option_size,  option_subtract, option_unset
    };
    int optionIndex;
    static char *options[] = {
	"add",          "array",      "create",        "exists",
	"foreach",      "get",        "merge",         "names",
	"set",          "size",       "subtract",      "unset",
	NULL
    };
    static Tcl_ObjCmdProc *commands[] = {
	/* Dictionary subcommands */
	(Tcl_ObjCmdProc *) DictionaryAddObjCmd,		/* add */
	(Tcl_ObjCmdProc *) DictionaryArrayObjCmd,	/* array */
	(Tcl_ObjCmdProc *) DictionaryCreateObjCmd,	/* create */
	(Tcl_ObjCmdProc *) DictionaryExistsObjCmd,	/* exists */
	(Tcl_ObjCmdProc *) DictionaryForeachObjCmd,	/* foreach */
	(Tcl_ObjCmdProc *) DictionaryGetObjCmd,		/* get */
	(Tcl_ObjCmdProc *) DictionaryMergeObjCmd,	/* merge */
	(Tcl_ObjCmdProc *) DictionaryNamesObjCmd,	/* names */
	(Tcl_ObjCmdProc *) DictionarySetObjCmd,		/* set */
	(Tcl_ObjCmdProc *) DictionarySizeObjCmd,	/* size */
	(Tcl_ObjCmdProc *) DictionarySubtractObjCmd,	/* subtract */
	(Tcl_ObjCmdProc *) DictionaryUnsetObjCmd	/* unset */
    };
    
    /*
     * Sanity check
     */

    if (objc < 2) {
	goto numberArgsError;
    }

    /*
     * Get the subcommand
     */

    if (Tcl_GetIndexFromObj(interp, objv[1], options, "option", 0, &optionIndex)
	== TCL_ERROR) {
	return TCL_ERROR;
    }

    /*
     * Subcommand calling
     */

    return (commands[optionIndex])(clientData, interp, objc, objv);
    
    /*
     * Error handling
     */

numberArgsError:
    Tcl_WrongNumArgs(interp, 1, objv, "option ?arg arg ...?");
    return TCL_ERROR;
}   


/*
 *----------------------------------------------------------------------
 *
 * Dictionary_Init --
 *
 *	This function is used to initialize the package. Here we check
 *	dependencies with other packages (eg Tcl), create new commands,
 *	and register our packages
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	New commands created, and package registered (see documentation).
 *
 *----------------------------------------------------------------------
 */

int Dictionary_Init(interp)
    Tcl_Interp *interp;
{  
    /* 
     * Dependencies 
     */

    if (Tcl_InitStubs(interp, TCL_VERSION, 0) == NULL)
	return TCL_ERROR;
    
    /* 
     * Initialization 
     */

    /* Query the Tcl list obj type */
    listObjTypePtr = Tcl_GetObjType ("list");

    /* Register our own dictionary type */
    Tcl_RegisterObjType(&dictionaryObjType);


    /* 
     * Commands creation 
     */

    Tcl_CreateObjCommand(interp, "::dictionary", 
				(Tcl_ObjCmdProc *) DictionaryObjCmd,
				(ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

    /* 
     * Package registration
     */

    if (Tcl_PkgProvide(interp, "dictionary", VERSION) == TCL_ERROR)
	return TCL_ERROR;

    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * Tclex_SafeInit --
 *
 *	This function is used to initialize the package in the special case
 *	of safe interpreters (like the Tcl Plugin). Here we check
 *	dependencies with other packages (eg Tcl), create new commands,
 *	and register our packages
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	New commands created, and package registered (see documentation).
 *
 *----------------------------------------------------------------------
 */

int Dictionary_SafeInit(interp)
    Tcl_Interp *interp;
{  
    /* For now, do the same as with unsafe interpreters */
    return Dictionary_Init(interp);
}

#ifdef WIN32
/*
 *----------------------------------------------------------------------
 *
 * DllEntryPoint --
 *
 *	This wrapper function is used by Windows to invoke the
 *	initialization code for the DLL.  If we are compiling
 *	with Visual C++, this routine will be renamed to DllMain.
 *	routine.
 *
 * Results:
 *	Returns TRUE;
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

BOOL APIENTRY
DllEntryPoint(hInst, reason, reserved)
    HINSTANCE hInst; /* Library instance handle. */
    DWORD reason;    /* Reason this function is being called. */
    LPVOID reserved; /* Not used. */
{
    return TRUE;
}
#endif /* WIN32 */

