/* 
 * dictionaryArray.c --
 *
 *	This module implements Tcl array interface to the dictionary
 *	object type. It consists of code ripped off Tcl source,
 *	in file generic/tclVar.c, especially array management
 *	procedures. These procedures were basically modified to
 *	take the dictionary object type into account as well as
 *	the existing list object type.
 *
 * Copyright (c) 1987-1994 The Regents of the University of California.
 * Copyright (c) 1994-1997 Sun Microsystems, Inc.
 * Copyright (c) 1998-1999 by Scriptics Corporation.
 * Copyright (c) 1999-2000 Frdric BONNET <frederic.bonnet@ciril.fr>
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <tclInt.h>

#include "dictionary.h"
#include "dictionaryInt.h"

static char *needArray =	"variable isn't array";

/*
 *----------------------------------------------------------------------
 *
 * CallTraces --
 *
 *	This procedure is invoked to find and invoke relevant
 *	trace procedures associated with a particular operation on
 *	a variable. This procedure invokes traces both on the
 *	variable and on its containing array (where relevant).
 *
 * Results:
 *	The return value is NULL if no trace procedures were invoked, or
 *	if all the invoked trace procedures returned successfully.
 *	The return value is non-NULL if a trace procedure returned an
 *	error (in this case no more trace procedures were invoked after
 *	the error was returned). In this case the return value is a
 *	pointer to a static string describing the error.
 *
 * Side effects:
 *	Almost anything can happen, depending on trace; this procedure
 *	itself doesn't have any side effects.
 *
 *----------------------------------------------------------------------
 */

static char *
CallTraces(iPtr, arrayPtr, varPtr, part1, part2, flags)
    Interp *iPtr;		/* Interpreter containing variable. */
    register Var *arrayPtr;	/* Pointer to array variable that contains
				 * the variable, or NULL if the variable
				 * isn't an element of an array. */
    Var *varPtr;		/* Variable whose traces are to be
				 * invoked. */
    char *part1, *part2;	/* Variable's two-part name. */
    int flags;			/* Flags passed to trace procedures:
				 * indicates what's happening to variable,
				 * plus other stuff like TCL_GLOBAL_ONLY,
				 * TCL_NAMESPACE_ONLY, and
				 * TCL_INTERP_DESTROYED. */
{
    register VarTrace *tracePtr;
    ActiveVarTrace active;
    char *result, *openParen, *p;
    Tcl_DString nameCopy;
    int copiedName;

    /*
     * If there are already similar trace procedures active for the
     * variable, don't call them again.
     */

    if (varPtr->flags & VAR_TRACE_ACTIVE) {
	return NULL;
    }
    varPtr->flags |= VAR_TRACE_ACTIVE;
    varPtr->refCount++;

    /*
     * If the variable name hasn't been parsed into array name and
     * element, do it here.  If there really is an array element,
     * make a copy of the original name so that NULLs can be
     * inserted into it to separate the names (can't modify the name
     * string in place, because the string might get used by the
     * callbacks we invoke).
     */

    copiedName = 0;
    if (part2 == NULL) {
	for (p = part1; *p ; p++) {
	    if (*p == '(') {
		openParen = p;
		do {
		    p++;
		} while (*p != '\0');
		p--;
		if (*p == ')') {
		    Tcl_DStringInit(&nameCopy);
		    Tcl_DStringAppend(&nameCopy, part1, (p-part1));
		    part2 = Tcl_DStringValue(&nameCopy)
			+ (openParen + 1 - part1);
		    part2[-1] = 0;
		    part1 = Tcl_DStringValue(&nameCopy);
		    copiedName = 1;
		}
		break;
	    }
	}
    }

    /*
     * Invoke traces on the array containing the variable, if relevant.
     */

    result = NULL;
    active.nextPtr = iPtr->activeTracePtr;
    iPtr->activeTracePtr = &active;
    if (arrayPtr != NULL) {
	arrayPtr->refCount++;
	active.varPtr = arrayPtr;
	for (tracePtr = arrayPtr->tracePtr;  tracePtr != NULL;
	     tracePtr = active.nextTracePtr) {
	    active.nextTracePtr = tracePtr->nextPtr;
	    if (!(tracePtr->flags & flags)) {
		continue;
	    }
	    result = (*tracePtr->traceProc)(tracePtr->clientData,
		    (Tcl_Interp *) iPtr, part1, part2, flags);
	    if (result != NULL) {
		if (flags & TCL_TRACE_UNSETS) {
		    result = NULL;
		} else {
		    goto done;
		}
	    }
	}
    }

    /*
     * Invoke traces on the variable itself.
     */

    if (flags & TCL_TRACE_UNSETS) {
	flags |= TCL_TRACE_DESTROYED;
    }
    active.varPtr = varPtr;
    for (tracePtr = varPtr->tracePtr; tracePtr != NULL;
	 tracePtr = active.nextTracePtr) {
	active.nextTracePtr = tracePtr->nextPtr;
	if (!(tracePtr->flags & flags)) {
	    continue;
	}
	result = (*tracePtr->traceProc)(tracePtr->clientData,
		(Tcl_Interp *) iPtr, part1, part2, flags);
	if (result != NULL) {
	    if (flags & TCL_TRACE_UNSETS) {
		result = NULL;
	    } else {
		goto done;
	    }
	}
    }

    /*
     * Restore the variable's flags, remove the record of our active
     * traces, and then return.
     */

    done:
    if (arrayPtr != NULL) {
	arrayPtr->refCount--;
    }
    if (copiedName) {
	Tcl_DStringFree(&nameCopy);
    }
    varPtr->flags &= ~VAR_TRACE_ACTIVE;
    varPtr->refCount--;
    iPtr->activeTracePtr = active.nextPtr;
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * VarErrMsg --
 *
 *      Generate a reasonable error message describing why a variable
 *      operation failed.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The interp's result is set to hold a message identifying the
 *      variable given by part1 and part2 and describing why the
 *      variable operation failed.
 *
 *----------------------------------------------------------------------
 */

static void
VarErrMsg(interp, part1, part2, operation, reason)
    Tcl_Interp *interp;         /* Interpreter in which to record message. */
    char *part1, *part2;        /* Variable's two-part name. */
    char *operation;            /* String describing operation that failed,
                                 * e.g. "read", "set", or "unset". */
    char *reason;               /* String describing why operation failed. */
{
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "can't ", operation, " \"", part1,
	    (char *) NULL);
    if (part2 != NULL) {
        Tcl_AppendResult(interp, "(", part2, ")", (char *) NULL);
    }
    Tcl_AppendResult(interp, "\": ", reason, (char *) NULL);
}

/*
 *----------------------------------------------------------------------
 *
 * TclArraySetDictionary --
 *
 *	Set the elements of an array from a dictionary.  If there are 
 *	no elements to set, create an empty array.
 *
 * Results:
 *	A standard Tcl result object.
 *
 * Side effects:
 *	A variable will be created if one does not already exist.
 *
 *----------------------------------------------------------------------
 */

int
TclArraySetDictionary(interp, arrayNameObj, arrayElemObj)
    Tcl_Interp *interp;		/* Current interpreter. */
    Tcl_Obj *arrayNameObj;	/* The array name. */
    Tcl_Obj *arrayElemObj;	/* The array elements dictionary.  If this is
				 * NULL, create an empty array. */
{
    Var *varPtr, *arrayPtr;
    Tcl_Obj **elemPtrs;
    int result, elemLen, i;
    char *varName, *p;
    
    varName = TclGetString(arrayNameObj);
    for (p = varName; *p ; p++) {
	if (*p == '(') {
	    do {
		p++;
	    } while (*p != '\0');
	    p--;
	    if (*p == ')') {
		VarErrMsg(interp, varName, NULL, "set", needArray);
		return TCL_ERROR;
	    }
	    break;
	}
    }

    varPtr = TclLookupVar(interp, varName, (char *) NULL, /*flags*/ 0,
            /*msg*/ 0, /*createPart1*/ 0, /*createPart2*/ 0, &arrayPtr);

    if (arrayElemObj != NULL) {
	if (arrayElemObj->typePtr == &dictionaryObjType) {
	    /* The specified element list is a dictionary */
	    Tcl_HashTable *tablePtr;
	    Tcl_HashSearch search;
	    Tcl_HashEntry *entryPtr;
	    Tcl_Obj *namePtr;

	    result = Dict_DictionaryObjGetElements(interp, arrayElemObj, &tablePtr);
	    if (result != TCL_OK) {
		return result;
	    }

	    for (entryPtr = Tcl_FirstHashEntry(tablePtr, &search);
		 entryPtr != (Tcl_HashEntry*) NULL;
		 entryPtr = Tcl_NextHashEntry(&search)) {

		namePtr = Tcl_NewStringObj(Tcl_GetHashKey(tablePtr, entryPtr), -1);
		if (Tcl_ObjSetVar2(interp, arrayNameObj,
				   namePtr, (Tcl_Obj *) Tcl_GetHashValue(entryPtr),
				   TCL_LEAVE_ERR_MSG) == NULL) {
		    Tcl_DecrRefCount(namePtr);
		    return TCL_ERROR;
		}
		Tcl_DecrRefCount(namePtr);
	    }
	    return TCL_OK;

	} else {
	    result = Tcl_ListObjGetElements(interp, arrayElemObj,
		    &elemLen, &elemPtrs);
	    if (result != TCL_OK) {
		return result;
	    }
	    if (elemLen & 1) {
		Tcl_ResetResult(interp);
		Tcl_AppendToObj(Tcl_GetObjResult(interp),
			"list must have an even number of elements", -1);
		return TCL_ERROR;
	    }
	    if (elemLen > 0) {
		for (i = 0;  i < elemLen;  i += 2) {
		    if (Tcl_ObjSetVar2(interp, arrayNameObj, elemPtrs[i],
			    elemPtrs[i+1], TCL_LEAVE_ERR_MSG) == NULL) {
			result = TCL_ERROR;
			break;
		    }
		}
		return result;
	    }
	}
    }
    
    /*
     * The list is empty make sure we have an array, or create
     * one if necessary.
     */
    
    if (varPtr != NULL) {
	if (!TclIsVarUndefined(varPtr) && TclIsVarArray(varPtr)) {
	    /*
	     * Already an array, done.
	     */
	    
	    return TCL_OK;
	}
	if (TclIsVarArrayElement(varPtr) || !TclIsVarUndefined(varPtr)) {
	    /*
	     * Either an array element, or a scalar: lose!
	     */
	    
	    VarErrMsg(interp, varName, (char *)NULL, "array set", needArray);
	    return TCL_ERROR;
	}
    } else {
	/*
	 * Create variable for new array.
	 */
	
	varPtr = TclLookupVar(interp, varName, (char *) NULL,
		TCL_LEAVE_ERR_MSG, "set",
	        /*createPart1*/ 1, /*createPart2*/ 0, &arrayPtr);

	/*
	 * Still couldn't do it - this can occur if a non-existent
	 * namespace was specified
	 */
	if (varPtr == NULL) {
	    return TCL_ERROR;
	}
    }
    TclSetVarArray(varPtr);
    TclClearVarUndefined(varPtr);
    varPtr->value.tablePtr =
	(Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(varPtr->value.tablePtr, TCL_STRING_KEYS);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * DictionaryArrayObjCmd --
 *
 *	This procedure is invoked to process the "array"
 *	subcommand for the Tcl command "dictionary". It either
 *	gets or sets arrays to or from a dictionary.
 *	Most of the code was taken from the Tcl array command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Objects type may change.
 *
 *--------------------------------------------------------------
 */

int
DictionaryArrayObjCmd(clientData, interp, objc, objv)
    ClientData	clientData;
    Tcl_Interp	*interp;
    int		objc;
    Tcl_Obj	*CONST objv[];
{
    static char *options[] = {
	"get",      "set",      "new",
	NULL
    };
    static enum {
	option_get, option_set, option_new
    };
    int optionIndex;

    Interp *iPtr = (Interp *) interp;
    Var *varPtr, *arrayPtr;
    Tcl_HashEntry *hPtr;
    int notArray;
    char *varName, *msg;
    int result;

    /* 
     * Sanity check
     */

    if (objc < 4) {
	goto numberArgsError;
    }

    /*
     * Get the subcommand
     */

    if (Tcl_GetIndexFromObj(interp, objv[2], options, "option", 0, &optionIndex) 
	== TCL_ERROR) {
	return TCL_ERROR;
    }

    /*
     * Locate the array variable (and it better be an array).
     */
    
    varName = TclGetString(objv[3]);
    varPtr = TclLookupVar(interp, varName, (char *) NULL, /*flags*/ 0,
            /*msg*/ 0, /*createPart1*/ 0, /*createPart2*/ 0, &arrayPtr);

    notArray = 0;
    if ((varPtr == NULL) || !TclIsVarArray(varPtr)
	    || TclIsVarUndefined(varPtr)) {
	notArray = 1;
    }

    /*
     * Special array trace used to keep the env array in sync for
     * array names, array get, etc.
     */

    if (varPtr != NULL && varPtr->tracePtr != NULL) {
	msg = CallTraces(iPtr, arrayPtr, varPtr, varName, NULL,
		(TCL_LEAVE_ERR_MSG|TCL_NAMESPACE_ONLY|TCL_GLOBAL_ONLY|
		TCL_TRACE_ARRAY));
	if (msg != NULL) {
	    VarErrMsg(interp, varName, NULL, "trace array", msg);
	    return TCL_ERROR;
	}
    }

    switch (optionIndex) {
    case option_get: {
	    Tcl_HashSearch search;
	    Var *varPtr2;
	    char *pattern = NULL;
	    char *name;
	    int newKey;
	    Tcl_Obj *namePtr, *valuePtr;
	    Tcl_Obj *resultPtr;
	    
	    if (objc != 4 && objc != 5) {
		goto getNumberArgsError;
	    }
	    if (notArray) {
	        return TCL_OK;
	    }

	    resultPtr =  Dict_NewDictionaryObj(0, NULL);
	    Tcl_IncrRefCount(resultPtr);

	    if (objc == 5) {
	        pattern = TclGetString(objv[4]);
	    }
	    for (hPtr = Tcl_FirstHashEntry(varPtr->value.tablePtr, &search);
		 hPtr != NULL;  hPtr = Tcl_NextHashEntry(&search)) {
	        varPtr2 = (Var *) Tcl_GetHashValue(hPtr);
		if (TclIsVarUndefined(varPtr2)) {
		    continue;
		}
		name = Tcl_GetHashKey(varPtr->value.tablePtr, hPtr);
		if ((objc == 5) && !Tcl_StringMatch(name, pattern)) {
		    continue;	/* element name doesn't match pattern */
		}
		
		namePtr = Tcl_NewStringObj(name, -1);
		
		valuePtr = Tcl_ObjGetVar2(interp, objv[3], namePtr,
	                TCL_LEAVE_ERR_MSG);
		if (valuePtr == NULL) {
		    Tcl_DecrRefCount(namePtr); /* free unneeded name obj */
		    Tcl_DecrRefCount(resultPtr);
		    return result;
		}
		result = Dict_DictionaryObjSetElement(interp, resultPtr,
			name, valuePtr, &newKey);
		if (result != TCL_OK) {
		    Tcl_DecrRefCount(namePtr); /* free unneeded name obj */
		    Tcl_DecrRefCount(resultPtr);
		    return result;
		}
	    }
	    
	    Tcl_SetObjResult(interp, resultPtr);
	    Tcl_DecrRefCount(resultPtr);
	    return TCL_OK;
	}

    case option_set:
    case option_new: {
	    if (objc != 5) {
		goto setNewNumberArgsError;
	    }
	    if (varPtr != NULL && optionIndex == option_new) {
		result = Tcl_UnsetVar(interp, varName, TCL_LEAVE_ERR_MSG);
		if (result != TCL_OK) {
		    return result;
		}
	    }
	    return(TclArraySetDictionary(interp, objv[3], objv[4]));
	}
    }

    /* 
     * Error handling
     */

numberArgsError:
    Tcl_WrongNumArgs(interp, 2, objv, "option arrayName ?args?");
    return TCL_ERROR;

getNumberArgsError:
    Tcl_WrongNumArgs(interp, 3, objv, "arrayName ?pattern?");
    return TCL_ERROR;

setNewNumberArgsError:
    Tcl_WrongNumArgs(interp, 3, objv, "arrayName dictionary");
    return TCL_ERROR;
}


