/* 
 * dictionaryObj.c --
 *
 *	This module implements the dictionary object internals. 
 *
 * Copyright (c) 1999-2000 Frdric BONNET <frederic.bonnet@ciril.fr>
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "dictionaryInt.h"

/*
 * Prototypes for procedures defined later in this file:
 */

static void		DupDictionaryInternalRep _ANSI_ARGS_((Tcl_Obj *srcPtr,
			    Tcl_Obj *copyPtr));
static void		FreeDictionaryInternalRep _ANSI_ARGS_((
			    Tcl_Obj *listPtr));
static int		SetDictionaryFromAny _ANSI_ARGS_((Tcl_Interp *interp,
			    Tcl_Obj *objPtr));
static void		UpdateStringOfDictionary _ANSI_ARGS_((
			    Tcl_Obj *listPtr));

/*
 * The structure below defines the dictionary Tcl object type by means of
 * procedures that can be invoked by generic object code.
 */

Tcl_ObjType dictionaryObjType = {
    "dictionary",			/* name */
    FreeDictionaryInternalRep,		/* freeIntRepProc */
    DupDictionaryInternalRep,		/* dupIntRepProc */
    UpdateStringOfDictionary,		/* updateStringProc */
    SetDictionaryFromAny		/* setFromAnyProc */
};


Tcl_Obj *
Dict_NewDictionaryObj(objc, objv)
    int objc;			/* Count of objects referenced by objv. */
    Tcl_Obj *CONST objv[];	/* An array of pointers to Tcl objects. */
{
    register Tcl_Obj *dictObj;
    register Dictionary *dictRepPtr;
    register Tcl_HashEntry *entryPtr;
    int i, newEntry;

    if (objc % 2 != 0) {
	/* Must have an even number of objects */
	return NULL;
    }

    dictObj = Tcl_NewObj();

    if (objc > 0) {
	Tcl_InvalidateStringRep(dictObj);

	/*
	 * Create and populate the elements table
	 */

	dictRepPtr = (Dictionary *) Tcl_Alloc(sizeof(Dictionary));
	Tcl_InitHashTable (&dictRepPtr->elements, TCL_STRING_KEYS);

	for (i=0; i < objc; i+=2) {
	    /* First create entry with even-numbered list element */
	    entryPtr = Tcl_CreateHashEntry(&dictRepPtr->elements,
					   Tcl_GetStringFromObj(objv[i], NULL),
					   &newEntry);

	    /* Then set its value with odd-numbered list element */
	    Tcl_IncrRefCount(objv[i+1]);
	    Tcl_SetHashValue(entryPtr, (ClientData) objv[i+1]);
	}

	/*
	 * Set object type
	 */

	dictObj->internalRep.otherValuePtr = (VOID *) dictRepPtr;
	dictObj->typePtr                   = &dictionaryObjType;
    }

    return dictObj;
}

int
Dict_DictionaryObjGetSize(interp, dictPtr, sizePtr)
    Tcl_Interp	*interp;	/* Used to report errors if not NULL. */
    Tcl_Obj	*dictPtr;	/* Dictionary object */
    int		*sizePtr;	/* Where to store the size of dictionary */
{
    register Dictionary *dictRepPtr;

    /*
     * We are expecting dictionary object
     */

    if (Tcl_ConvertToType(interp, dictPtr, &dictionaryObjType)
	== TCL_ERROR) {
        return TCL_ERROR;
    }

    dictRepPtr = (Dictionary *) dictPtr->internalRep.otherValuePtr;
    *sizePtr = dictRepPtr->elements.numEntries;
    return TCL_OK;
}

int
Dict_DictionaryObjGetElement(interp, dictPtr, key, objPtrPtr)
    Tcl_Interp	*interp;	/* Used to report errors if not NULL. */
    Tcl_Obj	*dictPtr;	/* Dictionary object */
    char	*key;		/* Element key */
    Tcl_Obj	**objPtrPtr;	/* The resulting Tcl_Obj* is stored here. */
{
    register Dictionary *dictRepPtr;
    Tcl_HashEntry *entryPtr;

    /*
     * We are expecting dictionary object
     */

    if (Tcl_ConvertToType(interp, dictPtr, &dictionaryObjType)
	== TCL_ERROR) {
        return TCL_ERROR;
    }

    dictRepPtr = (Dictionary *) dictPtr->internalRep.otherValuePtr;
    entryPtr = Tcl_FindHashEntry(&dictRepPtr->elements, key);
    if (!entryPtr) {
	*objPtrPtr = NULL;
	return TCL_OK;
    }

    *objPtrPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    return TCL_OK;
}

int
Dict_DictionaryObjSetElement(interp, dictPtr, key, valuePtr, newPtr)
    Tcl_Interp	*interp;	/* Used to report errors if not NULL. */
    Tcl_Obj	*dictPtr;	/* Dictionary object */
    char	*key;		/* Element key */
    Tcl_Obj	*valuePtr;	/* Element value */
    int		*newPtr;	/* Where to report element creation */
{
    register Dictionary *dictRepPtr;
    Tcl_HashEntry *entryPtr;
    Tcl_Obj *oldValuePtr = NULL;

    /*
     * We are expecting dictionary object
     */

    if (Tcl_ConvertToType(interp, dictPtr, &dictionaryObjType)
	== TCL_ERROR) {
        return TCL_ERROR;
    }

    dictRepPtr = (Dictionary *) dictPtr->internalRep.otherValuePtr;
    entryPtr = Tcl_CreateHashEntry(&dictRepPtr->elements, key, newPtr);
    if (!*newPtr) {
	oldValuePtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    }
    Tcl_SetHashValue(entryPtr, (ClientData) valuePtr);
    Tcl_IncrRefCount(valuePtr);
    if (oldValuePtr) {
	Tcl_DecrRefCount(oldValuePtr);
    }

    if (oldValuePtr != valuePtr) {
	/* Invalidate string rep */
	Tcl_InvalidateStringRep(dictPtr);
    }
    return TCL_OK;
}

int
Dict_DictionaryObjUnsetElement(interp, dictPtr, key)
    Tcl_Interp	*interp;	/* Used to report errors if not NULL. */
    Tcl_Obj	*dictPtr;	/* Dictionary object */
    char	*key;		/* Element key */
{
    register Dictionary *dictRepPtr;
    Tcl_HashEntry *entryPtr;

    /*
     * We are expecting dictionary object
     */

    if (Tcl_ConvertToType(interp, dictPtr, &dictionaryObjType)
	== TCL_ERROR) {
        return TCL_ERROR;
    }

    dictRepPtr = (Dictionary *) dictPtr->internalRep.otherValuePtr;
    entryPtr = Tcl_FindHashEntry(&dictRepPtr->elements, key);
    if (!entryPtr) {
	if (interp) {
	    Tcl_AppendResult(interp, 
			     "no such key: \"", key, "\"",
			     NULL);
	}
	return TCL_ERROR;
    }
    Tcl_DeleteHashEntry(entryPtr);

    /* Invalidate string rep */
    Tcl_InvalidateStringRep(dictPtr);

    return TCL_OK;
}

int
Dict_DictionaryObjGetElements(interp, dictPtr, tablePtrPtr)
    Tcl_Interp	*interp;		/* Used to report errors if not NULL. */
    Tcl_Obj	*dictPtr;		/* Dictionary object */
    Tcl_HashTable **tablePtrPtr;	/* Where to store table */
{
    register Dictionary *dictRepPtr;

    /*
     * We are expecting dictionary object
     */

    if (Tcl_ConvertToType(interp, dictPtr, &dictionaryObjType)
	== TCL_ERROR) {
        return TCL_ERROR;
    }

    dictRepPtr = (Dictionary *) dictPtr->internalRep.otherValuePtr;
    *tablePtrPtr = &dictRepPtr->elements;

    return TCL_OK;
}






/*
 *----------------------------------------------------------------------
 *
 * FreeDictionaryInternalRep --
 *
 *	Deallocate the storage associated with a dictionary object's 
 *	internal representation.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Frees setPtr's Set* internal representation and sets setPtr's
 *	internalRep.otherValuePtr to NULL. Decrements the ref counts
 *	of all element objects, which may free them.
 *
 *----------------------------------------------------------------------
 */

static void
FreeDictionaryInternalRep(dictPtr)
    Tcl_Obj *dictPtr;		/* Dictionary object with internal rep to free. */
{
    register Dictionary *dictRepPtr 
	= (Dictionary *) dictPtr->internalRep.otherValuePtr;
    Tcl_HashSearch search;
    Tcl_HashEntry *entryPtr;

    /* 
     * Free entry values 
     */

    for (entryPtr = Tcl_FirstHashEntry(&dictRepPtr->elements, &search);
	 entryPtr != (Tcl_HashEntry*) NULL;
	 entryPtr = Tcl_NextHashEntry(&search)) {

	Tcl_DecrRefCount((Tcl_Obj *) Tcl_GetHashValue(entryPtr));
    }

    Tcl_DeleteHashTable(&dictRepPtr->elements);
    Tcl_Free((char *) dictRepPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * DupDictionaryInternalRep --
 *
 *	Initialize the internal representation of a dictionary Tcl_Obj to a
 *	copy of the internal representation of an existing dictionary object. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	"srcPtr"s set internal rep pointer should not be NULL and we assume
 *	it is not NULL. We set "copyPtr"s internal rep to a pointer to a
 *	newly allocated Dictionary structure whose elements, in turn, 
 *	contains "srcPtr"s element objects. Those element objects are not 
 *	actually copied but are shared between "srcPtr" and "copyPtr". 
 *	The ref count of each element object is incremented.
 *
 *----------------------------------------------------------------------
 */

static void
DupDictionaryInternalRep(srcPtr, copyPtr)
    Tcl_Obj *srcPtr;		/* Object with internal rep to copy. */
    Tcl_Obj *copyPtr;		/* Object with internal rep to set. */
{
    register Dictionary*  srcDictRepPtr 
	= (Dictionary *) srcPtr->internalRep.otherValuePtr;
    register Dictionary*  copyDictRepPtr;
    Tcl_HashEntry *entrySrcPtr, *entryCopyPtr;
    Tcl_HashSearch search;
    int newEntry;
    Tcl_Obj *valuePtr;

    /*
     * Allocate a new Dictionary structure and initialize it with a copy of 
     * the source's elements hashtable.
     */

    /* 
     * TODO: check if low-level hashtable copy facilities can be used.
     * Rather than copying tables one element at once, maybe a raw copy
     * of the whole table could be done? 
     */

    copyDictRepPtr = (Dictionary *) Tcl_Alloc(sizeof(Dictionary));
    Tcl_InitHashTable (&copyDictRepPtr->elements, TCL_STRING_KEYS);

    for (entrySrcPtr = Tcl_FirstHashEntry(&srcDictRepPtr->elements, &search);
	 entrySrcPtr != (Tcl_HashEntry*) NULL;
	 entrySrcPtr = Tcl_NextHashEntry(&search)) {

	/* Create a new entry in copy hash table */
	entryCopyPtr = Tcl_CreateHashEntry(&copyDictRepPtr->elements, 
					   Tcl_GetHashKey(&copyDictRepPtr->elements, entrySrcPtr), 
					   &newEntry);

	/* Set the value of the newly created entry */
	valuePtr = (Tcl_Obj *) Tcl_GetHashValue(entrySrcPtr);
	Tcl_IncrRefCount(valuePtr);
	Tcl_SetHashValue(entryCopyPtr, (ClientData) valuePtr);
    }
    
    copyPtr->internalRep.otherValuePtr = (VOID *) copyDictRepPtr;
    copyPtr->typePtr                   = &dictionaryObjType;
}

/*
 *----------------------------------------------------------------------
 *
 * SetDictionaryFromAny --
 *
 *	Attempt to generate a dictionary internal form for the Tcl object
 *	"objPtr".
 *
 * Results:
 *	The return value is TCL_OK or TCL_ERROR. If an error occurs during
 *	conversion, an error message is left in the interpreter's result
 *	unless "interp" is NULL.
 *
 * Side effects:
 *	If no error occurs, a dictionary is stored as "objPtr"s internal
 *	representation.
 *
 *----------------------------------------------------------------------
 */

static int
SetDictionaryFromAny(interp, objPtr)
    Tcl_Interp *interp;		/* Used for error reporting if not NULL. */
    Tcl_Obj *objPtr;		/* The object to convert. */
{
    Tcl_ObjType* oldTypePtr = objPtr->typePtr;
    int listc, i, newEntry;
    Tcl_Obj** listv;
    register Dictionary* dictRepPtr;
    Tcl_HashEntry *entryPtr;

    /*
     * First convert object to a list, then iterate over list elements.
     */

    /*
     * TODO: avoid the extra conversion to a list object when the object
     * is not already a list. Else, iterate over the string rep. Grab
     * the code from the list object type.
     */
    
    if (Tcl_ConvertToType(interp, objPtr, listObjTypePtr)
	== TCL_ERROR) {
	/* objPtr may not have a valid list rep. */
        return TCL_ERROR;
    }

    if (Tcl_ListObjGetElements(interp, objPtr, &listc, &listv)
	== TCL_ERROR) {
        return TCL_ERROR;
    }

    /* List must have an even number of elements */
    if (listc % 2 != 0) {
	Tcl_AppendResult(interp, 
			 "list must have an even number of elements", 
			 NULL);
	return TCL_ERROR;
    }

    /*
     * Create and populate the elements table
     */

    dictRepPtr = (Dictionary *) Tcl_Alloc(sizeof(Dictionary));
    Tcl_InitHashTable (&dictRepPtr->elements, TCL_STRING_KEYS);

    for (i=0; i < listc; i+=2) {
	/* First create entry with even-numbered list element */
        entryPtr = Tcl_CreateHashEntry(&dictRepPtr->elements,
				       Tcl_GetStringFromObj(listv[i], NULL),
				       &newEntry);

	/* Then set its value with odd-numbered list element */
	Tcl_IncrRefCount(listv[i+1]);
	Tcl_SetHashValue(entryPtr, (ClientData) listv[i+1]);
    }

    /*
     * Free the old internalRep before setting the new one. We do this as
     * late as possible to allow the conversion code, in particular
     * Tcl_GetStringFromObj, to use that old internalRep.
     */

    if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
	oldTypePtr->freeIntRepProc(objPtr);
    }

    objPtr->internalRep.otherValuePtr = (VOID *) dictRepPtr;
    objPtr->typePtr                   = &dictionaryObjType;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * UpdateStringOfDictionary --
 *
 *	Update the string representation for a dictionary object.
 *	Note: This procedure does not invalidate an existing old string rep
 *	so storage will be lost if this has not already been done. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The object's string is set to a valid string that results from
 *	the dictionary-to-string conversion. This string will be empty if the
 *	dictionary has no elements. The dictionary internal representation
 *	should not be NULL and we assume it is not NULL.
 *
 *----------------------------------------------------------------------
 */

static void
UpdateStringOfDictionary(dictPtr)
    Tcl_Obj *dictPtr;		/* Dictionary object with string rep to update. */
{
#   define LOCAL_SIZE 20
    int localFlags[LOCAL_SIZE], *flagPtr;
    Dictionary *dictRepPtr = (Dictionary *) dictPtr->internalRep.otherValuePtr;
    int numElems = dictRepPtr->elements.numEntries * 2;
    register int i;
    char *key, *value, *dst;
    int valueLength;
    Tcl_HashEntry* entryPtr;
    Tcl_HashSearch search;

    /*
     * Convert each element of the set to string form and then convert it
     * to proper dictionary element form, adding it to the result buffer.
     */

    /*
     * Pass 1: estimate space, gather flags.
     */

    if (numElems <= LOCAL_SIZE) {
	flagPtr = localFlags;
    } else {
	flagPtr = (int *) Tcl_Alloc((unsigned) numElems*sizeof(int));
    }
    dictPtr->length = 1;
    for (entryPtr = Tcl_FirstHashEntry(&dictRepPtr->elements, &search), i=0;
	 entryPtr != (Tcl_HashEntry*) NULL;
	 entryPtr = Tcl_NextHashEntry(&search), i+=2) {

        key = Tcl_GetHashKey(&dictRepPtr->elements, entryPtr);
	dictPtr->length += Tcl_ScanElement(key, &flagPtr[i]) + 1;

	value = Tcl_GetStringFromObj((Tcl_Obj *) Tcl_GetHashValue(entryPtr),
				     &valueLength);
	dictPtr->length += Tcl_ScanCountedElement(value, valueLength, 
						  &flagPtr[i+1]) + 1;
    }

    /*
     * Pass 2: copy into string rep buffer.
     */

    dictPtr->bytes = Tcl_Alloc((unsigned) dictPtr->length);
    dst = dictPtr->bytes;
    for (entryPtr = Tcl_FirstHashEntry (&dictRepPtr->elements, &search), i=0;
	 entryPtr != (Tcl_HashEntry*) NULL;
	 entryPtr = Tcl_NextHashEntry (&search), i+=2) {

        key = Tcl_GetHashKey(&dictRepPtr->elements, entryPtr);
	dst += Tcl_ConvertElement(key, dst, flagPtr[i]);
	*dst = ' ';
	dst++;

	value = Tcl_GetStringFromObj((Tcl_Obj *) Tcl_GetHashValue(entryPtr),
				     &valueLength);
	dst += Tcl_ConvertCountedElement(value, valueLength, dst, 
					 flagPtr[i+1]);
	*dst = ' ';
	dst++;
    }
    if (flagPtr != localFlags) {
	Tcl_Free((char *) flagPtr);
    }
    if (dst == dictPtr->bytes) {
	*dst = 0;
    } else {
	dst--;
	*dst = 0;
    }
    dictPtr->length = dst - dictPtr->bytes;
}

